unit hiTetris;

interface

uses Windows, Kol, Share, Debug, Win, Messages;

type

  TBig = array[0..0] of Integer;

  TPanelColor = record 
    R1, G1, B1, R2, G2, B2: Byte; 
  end;

  TPixelToClear = record 
    mX, mY: ShortInt; 
    R1, G1, B1: Byte; 
    cTop, cLeft: Word; 
  end;

  TObjectColor = record 
    Top, Left: ShortInt; 
    Height, Width: Byte; 
    PanelColor: TPanelColor; 
    PanelArray: array of array of Boolean;
    Visible: Boolean;
  end;

  TFormPanelColor = record 
    Visible: Boolean; 
    PanelColor: TPanelColor; 
  end;

  THITetris = class(THIWin)
    procedure Init; override;
    destructor Destroy; override;
    procedure TimerFPSTimer(Sender: PObj);
    
    procedure FormPaint(Sender: PControl; DC: HDC);
    
    procedure _onKeyDown(Sender: PControl; var Key: Longint; Shift: DWORD); override;
    procedure TimerGravityTimer(Sender: PObj);
  private
    
    TimerFPS: PTimer; 
    TimerGravity: PTimer; 
    FScr: Pointer; 
    FScrBitmap: Kol.PBitmap; 
    FCanvaGamePanel: array of array of TColor; 
    
    FGameOver: Boolean; 
    FCurrGamePause: Boolean; 
    FCurrGameLevel: Byte; 
    FCurrGameLines: Byte; 
    FPanelBlikStep: Byte; 
    FCurScore, FCurLevel, FCurLines: string[8];
    
    ObjectColorCurr, 
      ObjectColorBuff, 
      ObjectColorCapt, 
      ObjectColorShad, 
      ObjectColorNext: TObjectColor; 
    
    FormPanelColor: array of array of TFormPanelColor;
    
    PixelToClear: array of array of array of array of TPixelToClear;
    yNumLines: Byte; 
    IsInit: boolean; 
  public
    
    function IncScore(n: Word; InScore: string): string; 
    procedure CreateScrBitMap(BitMapH, BitMapW: Integer); 
    procedure RefreshBackGround; 
    procedure PaintGrid; 
    procedure FillBackGround; 
    procedure SetLengthCanvaGamePanel; 
    procedure PaintingAllPanels; 
    procedure CreateNextObjectColor(NumObj: Byte); 
    procedure PaintClockFace(sNumber: string; cLeft, cHeight: Word; R, G, B: Byte; VisibleZero: Boolean); 
    procedure WriteText; 
    procedure SetLengthFormPanelColor; 
    procedure MoweObjectColor(mw: Byte); 
    procedure ReStartGame; 
    procedure ChangeObjectColor; 
    procedure TurnObjectColor(angle: ShortInt); 
    procedure EraseLinePanel; 
    function DropObject(MyObjectColor: TObjectColor): TObjectColor; 
    procedure ReShowScore; 
    procedure CalculateLevel(SumLines: Word);
  end;

const
  PanelColorH = 24; 
  PanelColorW = 24; 
  PanelColorSummH = 20; 
  PanelColorSummW = 10; 
  FrameDepth = 20; 
  PanelGameH = PanelColorH * PanelColorSummH + PanelColorSummH; 
  PanelGameW = PanelColorW * PanelColorSummW + PanelColorSummW; 
  PanelInfoH = PanelGameH; 
  PanelInfoW = (PanelColorW + 1) * 6; 
  FormH = PanelGameH + FrameDepth * 2; 
  FormW = PanelGameW + FrameDepth * 4 + PanelInfoW; 
  StepSplashingH = 3; 
  StepSplashingW = StepSplashingH;

implementation

destructor THITetris.Destroy;
begin
{$IFNDEF F_P}
  TimerFPS.Free;
  TimerGravity.Free;
{$ENDIF}
  FScrBitmap.Free;
  SetLength(FormPanelColor, 0, 0);
  inherited Destroy;
end;

function THITetris.IncScore(n: Word; InScore: string): string;
var
  MyScore: Word;
  i: Byte;
  OutScore: string[8];
begin
  MyScore := Str2Int(InScore);
  Inc(MyScore, n);
  OutScore := Int2Str(MyScore);
  for i := Length(OutScore) to 7 do
    OutScore := '0' + OutScore;
  Result := OutScore;
end;

function THITetris.DropObject(MyObjectColor: TObjectColor): TObjectColor;
var
  x, y: Byte;
  cX, cY: ShortInt;
  CanMowe: Boolean;
label NextMoweObjectColor;
begin
  NextMoweObjectColor: 
  CanMowe := True; 
  with MyObjectColor do
    
    if Visible then begin
      for x := 0 to Width - 1 do begin
        
        if not CanMowe then
          Break;
        
        for y := 0 to Height - 1 do begin
          
          cX := Left + x;
          cY := Top + y + 1;
          
          if PanelArray[x, y] = True then begin
            if cY < 0 then
              cY := 0;
            if (cX < 0) or (cX > PanelColorSummW - 1) then
              cX := 0;
            if (cY > PanelColorSummH - 1) or
              (FormPanelColor[cX, cY].Visible) then
            begin
              CanMowe := False;
              Break;
            end;
          end;
        end;
      end;
    end;
  
  if CanMowe then
  begin
    with MyObjectColor do
      if Visible then
        
        inc(Top);
    
    goto NextMoweObjectColor;
  end;

  Result := MyObjectColor;
end;

procedure THITetris.CalculateLevel(SumLines: Word);
var
  SumLinesToNextLVL: Word;
  i: Byte;
begin
  SumLinesToNextLVL := 10;
  FCurLevel := '00000001';
  i := 1;
  while SumLinesToNextLVL < SumLines do
  begin
    inc(i);
    SumLinesToNextLVL := SumLinesToNextLVL + i * 10;
    FCurLevel := IncScore(1, FCurLevel);
  end;
end;

procedure THITetris.EraseLinePanel;
var
  x, y, Ymy: Byte;
  iW, iH, pixelR, pixelG, pixelB, cLvl: Byte;
  cLeft, cTop: Word;
  yErase: array of Byte;
  FindLine: Boolean;
  StepH, StepW: ShortInt;
begin

  yNumLines := 0;
  
  for y := 0 to PanelColorSummH - 1 do
  begin
    
    FindLine := True;
    for x := 0 to PanelColorSummW - 1 do
    begin
      if not FormPanelColor[x, y].Visible then
      begin
        FindLine := False;
        Break;
      end;
    end;
    
    if FindLine = True then
    begin
      inc(yNumLines); 
      FCurScore := IncScore(200, FCurScore); 
      SetLength(yErase, yNumLines);
      yErase[yNumLines - 1] := y; 
    end;
  end;
  
  if yNumLines = 0 then
    Exit;

  FCurLines := IncScore(yNumLines, FCurLines); 
  
  CalculateLevel(Str2Int(FCurLines));
  TimerGravity.Interval := 800 - (Str2Int(FCurLevel) - 1) * 40;
  
  for x := 0 to PanelColorSummW - 1 do
  begin
    cLeft := FrameDepth + PanelColorW * x + x + 1;
    for y := yNumLines - 1 downto 0 do
    begin
      yMy := yErase[y];
      cTop := FrameDepth + PanelColorH * yMy + yMy + 1;
      for iW := 0 to PanelColorW - 1 do begin
        
        StepW := ((iW + 1) - (PanelColorW div 2)) div StepSplashingW;
        
        for iH := 0 to PanelColorH - 1 do
        begin
          
          StepH := (iH + 1 - (PanelColorH div 2)) div StepSplashingH;
          
          PixelR := GetRValue(FScrBitmap.Canvas.Pixels[cLeft + iW + 1, cTop + iH + 1]);
          PixelG := GetGValue(FScrBitmap.Canvas.Pixels[cLeft + iW + 1, cTop + iH + 1]);
          PixelB := GetBValue(FScrBitmap.Canvas.Pixels[cLeft + iW + 1, cTop + iH + 1]);
          
          PixelToClear[x, y, iW, iH].mX := StepW * 2;
          PixelToClear[x, y, iW, iH].mY := StepH - 6;

          PixelToClear[x, y, iW, iH].R1 := PixelR;
          PixelToClear[x, y, iW, iH].G1 := PixelG;
          PixelToClear[x, y, iW, iH].B1 := PixelB;
          PixelToClear[x, y, iW, iH].cTop := cTop + iH;
          PixelToClear[x, y, iW, iH].cLeft := cLeft + iW;
        end;
      end;
      
      FormPanelColor[x, yMy].Visible := False;
    end;
  end;
  
  for x := 0 to PanelColorSummW - 1 do
  begin
    cLeft := FrameDepth + PanelColorW * x + x + 1;
    for y := yErase[yNumLines - 1] downto yNumLines - 1 do
    begin
      if FormPanelColor[x, y - yNumLines].Visible then begin
        FormPanelColor[x, y] := FormPanelColor[x, y - yNumLines];
        FormPanelColor[x, y - yNumLines].Visible := False;
      end;
    end;
  end;

end;

procedure THITetris.TurnObjectColor(angle: ShortInt);

  procedure TurnObject(b: Boolean);
  var
    x, y, i, iMax, iKol: Byte;
    cX, cY: ShortInt;
    CanMowe: Boolean;
    cXY, cXYb: array of boolean;
  begin
    if not ObjectColorCurr.Visible then 
      Exit;
    
    CanMowe := True;
    
    ObjectColorBuff := ObjectColorCurr;
    
    ObjectColorBuff.Height := ObjectColorCurr.Width;
    ObjectColorBuff.Width := ObjectColorCurr.Height;
    SetLength(ObjectColorBuff.PanelArray, ObjectColorBuff.Width, ObjectColorBuff.Height);
    
    ObjectColorBuff.Top := ObjectColorCurr.Top + ObjectColorCurr.Height - ObjectColorBuff.Height;
    ObjectColorBuff.Left := ObjectColorCurr.Left + ObjectColorCurr.Width - ObjectColorBuff.Width;
    
    with ObjectColorCurr do
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          if b then 
            ObjectColorBuff.PanelArray[y, (Width - 1) - x] := ObjectColorCurr.PanelArray[x, y]
          else 
            ObjectColorBuff.PanelArray[(Height - 1) - y, x] := ObjectColorCurr.PanelArray[x, y];
    
    with ObjectColorBuff do
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          
          if PanelArray[x, y] = True then begin
            cX := Left + x;
            cY := Top + y;
            
            if cY < 0 then
              cY := 0;
            
            if (cX < 0) or (cX > PanelColorSummW - 1) then
            begin
              CanMowe := False;
              break;
            end;
            if FormPanelColor[cX, cY].Visible then
            begin
              CanMowe := False;
              break;
            end;
          end;
    
    if CanMowe then 
      ObjectColorCurr := ObjectColorBuff;
  end;

begin
  case angle of
    90: TurnObject(False);
    -90: TurnObject(True);
  end;
  MoweObjectColor(5); 
end;

procedure THITetris.MoweObjectColor(mw: Byte);

  procedure FreezeObject; 
  var
    x, y: Byte;
  begin
    for x := 0 to ObjectColorCurr.Width - 1 do
      for y := 0 to ObjectColorCurr.Height - 1 do
        
        if ObjectColorCurr.PanelArray[x, y] = True then
          with FormPanelColor[x + ObjectColorCurr.Left, y + ObjectColorCurr.Top] do
          begin
            Visible := True;
            with PanelColor do
            begin
              R1 := ObjectColorCurr.PanelColor.R1;
              G1 := ObjectColorCurr.PanelColor.G1;
              B1 := ObjectColorCurr.PanelColor.B1;
              R2 := ObjectColorCurr.PanelColor.R2;
              G2 := ObjectColorCurr.PanelColor.G2;
              B2 := ObjectColorCurr.PanelColor.B2;
            end;
          end;
    
    ObjectColorCurr := ObjectColorNext;
    
    CreateNextObjectColor(random(7));
    
    EraseLinePanel;
  end;

  procedure MoweDownObject;
  var
    x, y: Byte;
    cX, cY: ShortInt;
    CanMowe: Boolean;
  begin
    CanMowe := True; 
    with ObjectColorCurr do
      
      if Visible then
        for x := 0 to Width - 1 do
          for y := 0 to Height - 1 do
            
            if PanelArray[x, y] = True then begin
              cX := Left + x;
              cY := Top + y + 1;
              if cY < 0 then
                cY := 0;
              if (cX < 0) or (cX > PanelColorSummW - 1) then
                cX := 0;
              if cY > PanelColorSummH - 1 then
              begin
                CanMowe := False;
                break;
              end;
              if FormPanelColor[cX, cY].Visible then
              begin
                
                CanMowe := False;
                
                if cY = 0 then
                  FGameOver := True;
                break;
              end;
            end;
    
    if ObjectColorCurr.Visible and not FGameOver then
      if CanMowe then
      begin
        
        with ObjectColorCurr do
          inc(Top);
      end else
      begin
        
        FreezeObject;
      end;
  end;

  procedure MoweLeftOrRightObject(n: ShortInt);
  var
    x, y: Byte;
    cX, cY: ShortInt;
    CanMowe: Boolean;
  begin
    CanMowe := True; 
    with ObjectColorCurr do
      
      if Visible then
        for x := 0 to Width - 1 do
          for y := 0 to Height - 1 do
            
            if PanelArray[x, y] = True then begin
              cX := Left + x + n;
              cY := Top + y;
              
              if cY < 0 then begin
                cY := 0;
                
              end;
              
              if (cX < 0) or (cX > PanelColorSummW - 1) then
              begin
                CanMowe := False;
                break;
              end;
              if FormPanelColor[cX, cY].Visible then
              begin
                CanMowe := False;
                break;
              end;
            end;
    
    if CanMowe and ObjectColorCurr.Visible then
    begin
      
      with ObjectColorCurr do
        inc(Left, n);
    end;
  end;

  procedure DroppingObject;
  begin
    
    ObjectColorBuff := DropObject(ObjectColorCurr);
    
    FCurScore := IncScore(ObjectColorBuff.Top - ObjectColorCurr.Top, FCurScore);
    
    ObjectColorCurr := ObjectColorBuff;
    
    FreezeObject;
  end;

  procedure ShowShandow;
  begin
    ObjectColorShad := ObjectColorCurr;
    
    ObjectColorShad := DropObject(ObjectColorShad);
  end;

begin

  case mw of
    0: MoweDownObject; 
    1: MoweLeftOrRightObject(-1); 
    2: MoweLeftOrRightObject(1); 
    3: MoweDownObject; 
    4: DroppingObject; 
  end;

  ShowShandow; 

end;

procedure THITetris.CreateNextObjectColor(NumObj: Byte);

  procedure ClearArray(cWidth, cHeight: Byte);
  var
    x, y: Byte;
  begin
    for x := 0 to cWidth - 1 do
      for y := 0 to cHeight - 1 do
        ObjectColorNext.PanelArray[x, y] := False;
  end;

begin
  with ObjectColorNext do begin
    Visible := True;
    case NumObj of
      0:
        begin
        
          Height := 2; Top := -Height;
          Width := 3; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          ClearArray(Width, Height);
        
          PanelArray[0, 0] := True;
          PanelArray[1, 0] := True;
          PanelArray[1, 1] := True;
          PanelArray[2, 1] := True;
        
          with PanelColor do begin
            R1 := 100; R2 := 0; 
            G1 := 143; G2 := 43; 
            B1 := 121; B2 := 21; 
          end;
        end;
      1:
        begin
        
          Height := 2; Top := -Height;
          Width := 3; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          ClearArray(Width, Height);
        
          PanelArray[0, 1] := True;
          PanelArray[1, 1] := True;
          PanelArray[1, 0] := True;
          PanelArray[2, 0] := True;
        
          with PanelColor do begin
            R1 := 104; R2 := 4; 
            G1 := 166; G2 := 66; 
            B1 := 117; B2 := 17; 
          end;
        end;
      2:
        begin
        
          Height := 2; Top := -Height;
          Width := 3; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          ClearArray(Width, Height);
        
          PanelArray[1, 0] := True;
          PanelArray[0, 1] := True;
          PanelArray[1, 1] := True;
          PanelArray[2, 1] := True;
        
          with PanelColor do begin
            R1 := 237; R2 := 138; 
            G1 := 128; G2 := 28; 
            B1 := 106; B2 := 6; 
          end;
        end;
      3:
        begin
        
          Height := 2; Top := -Height;
          Width := 2; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          PanelArray[0, 0] := True;
          PanelArray[1, 0] := True;
          PanelArray[0, 1] := True;
          PanelArray[1, 1] := True;
        
          with PanelColor do begin
            R1 := 163; R2 := 63; 
            G1 := 133; G2 := 33; 
            B1 := 135; B2 := 35; 
          end;
        end;
      4:
        begin
        
          Height := 2; Top := -Height;
          Width := 3; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          ClearArray(Width, Height);
        
          PanelArray[0, 0] := True;
          PanelArray[1, 0] := True;
          PanelArray[2, 0] := True;
          PanelArray[2, 1] := True;
        
          with PanelColor do begin
            R1 := 160; R2 := 60; 
            G1 := 101; G2 := 1; 
            B1 := 225; B2 := 125; 
          end;
        end;
      5:
        begin
        
          Height := 2; Top := -Height;
          Width := 3; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          ClearArray(Width, Height);
        
          PanelArray[0, 0] := True;
          PanelArray[1, 0] := True;
          PanelArray[2, 0] := True;
          PanelArray[0, 1] := True;
        
          with PanelColor do begin
            R1 := 206; R2 := 106; 
            G1 := 110; G2 := 10; 
            B1 := 252; B2 := 152; 
          end;
        end;
      6:
        begin
        
          Height := 1; Top := -Height;
          Width := 4; Left := (PanelColorSummW - Width) div 2;
          SetLength(PanelArray, Width, Height);
        
          PanelArray[0, 0] := True;
          PanelArray[1, 0] := True;
          PanelArray[2, 0] := True;
          PanelArray[3, 0] := True;
        
          with PanelColor do begin
            R1 := 137; R2 := 37; 
            G1 := 122; G2 := 22; 
            B1 := 234; B2 := 134; 
          end;
        end;
    end;
  end;
end;

procedure THITetris.PaintingAllPanels;

  procedure PaintingPanel(R1, G1, B1, R2, G2, B2: Byte; cLeft, cTop, cHeight, cWidth: Word);
  var
    y: integer;
  begin
    
    for y := 0 to cHeight do
      with FScrBitmap.Canvas{$IFNDEF F_P}^{$ENDIF} do
      begin
        Pen.Color := RGB(R1 - (R1 - R2) div cHeight * y,
          G1 - (G1 - G2) div cHeight * y,
          B1 - (B1 - B2) div cHeight * y);
        MoveTo(cLeft + 1, cTop + y + 1);
        LineTo(cLeft + cWidth + 1, cTop + y + 1);
      end;
    
    FScrBitmap.Canvas.Pen.Color := RGB(R2, G2, B2);
    FScrBitmap.Canvas.Moveto(cLeft, cTop + 1);
    FScrBitmap.Canvas.LineTo(cLeft + 1, cTop);
    FScrBitmap.Canvas.LineTo(cLeft + PanelColorW - 2, cTop);
    FScrBitmap.Canvas.LineTo(cLeft + PanelColorW - 1, cTop + 1);
    FScrBitmap.Canvas.LineTo(cLeft + PanelColorW - 1, cTop + PanelColorH - 2);
    FScrBitmap.Canvas.LineTo(cLeft + PanelColorW - 2, cTop + PanelColorH - 1);
    FScrBitmap.Canvas.LineTo(cLeft + 1, cTop + PanelColorH - 1);
    FScrBitmap.Canvas.LineTo(cLeft, cTop + PanelColorH - 2);
    FScrBitmap.Canvas.LineTo(cLeft, cTop + 1);
  end;

  procedure PaintingPixel(R1, G1, B1: Byte; cLeft, cTop: Word);
  begin
    with FScrBitmap.Canvas{$IFNDEF F_P}^{$ENDIF} do
      Pixels[cLeft, cTop] := RGB(R1, G1, B1);
  end;

var
  x, y, iW, iH: Byte;
  cTop, cLeft: integer;
begin
  
  for x := 0 to PanelColorSummW - 1 do
    for y := 0 to PanelColorSummH - 1 do
      if FormPanelColor[x, y].Visible then
      begin
        
        cLeft := FrameDepth + PanelColorW * x + x;
        cTop := FrameDepth + PanelColorH * y + y;
        
        with FormPanelColor[x, y].PanelColor do
          PaintingPanel(R1, G1, B1, R2, G2, B2,
            cLeft, cTop,
            PanelColorH - 3, PanelColorW - 2);
      end;
  
  cTop := FrameDepth + (PanelColorH + 1) * 9;
  cLeft := FrameDepth * 3 + (PanelColorW + 1) * (PanelColorSummW + 1);
  with ObjectColorNext do
    if ObjectColorNext.Visible then 
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          if PanelArray[x, y] = True then
            with PanelColor do
              PaintingPanel(R1, G1, B1, R2, G2, B2,
                cLeft + (x + (4 - ObjectColorNext.Width) div 2) * (PanelColorW + 1),
                cTop + (y + (4 - ObjectColorNext.Height) div 2) * (PanelColorH + 1),
                PanelColorH - 3, PanelColorW - 2);
  
  cTop := FrameDepth + (PanelColorH + 1) * 15;
  cLeft := FrameDepth * 3 + (PanelColorW + 1) * (PanelColorSummW + 1);
  with ObjectColorCapt do
    if ObjectColorCapt.Visible then 
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          if PanelArray[x, y] = True then
            with PanelColor do
              PaintingPanel(R1, G1, B1, R2, G2, B2,
                cLeft + (x + (4 - ObjectColorCapt.Width) div 2) * (PanelColorW + 1),
                cTop + (y + (4 - ObjectColorCapt.Height) div 2) * (PanelColorH + 1),
                PanelColorH - 3, PanelColorW - 2);
  
  cTop := FrameDepth + (PanelColorH + 1) * ObjectColorShad.Top;
  cLeft := FrameDepth + (PanelColorW + 1) * ObjectColorShad.Left;
  with ObjectColorShad do
    if Visible then 
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          if PanelArray[x, y] = True then
            with PanelColor do 
              PaintingPanel(R1 - (R1 - 120) div 2,
                G1 - (G1 - 120) div 2,
                B1 - (B1 - 120) div 2,
                R2 - (R2 - 120) div 2,
                G2 - (G2 - 120) div 2,
                B2 - (B2 - 120) div 2,
                cLeft + x * (PanelColorW + 1),
                cTop + y * (PanelColorH + 1),
                PanelColorH - 3, PanelColorW - 2);
  
  cTop := FrameDepth + (PanelColorH + 1) * ObjectColorCurr.Top;
  cLeft := FrameDepth + (PanelColorW + 1) * ObjectColorCurr.Left;
  with ObjectColorCurr do
    if Visible then 
      for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
          if PanelArray[x, y] = True then
            with PanelColor do
              PaintingPanel(R1, G1, B1, R2, G2, B2,
                cLeft + x * (PanelColorW + 1),
                cTop + y * (PanelColorH + 1),
                PanelColorH - 3, PanelColorW - 2);
  
  if yNumLines > 0 then
    for x := 0 to PanelColorSummW - 1 do
    begin
      for y := yNumLines - 1 downto 0 do
      begin
        for iW := 0 to PanelColorW - 1 do
        begin
          for iH := 0 to PanelColorH - 1 do
          begin
            with PixelToClear[x, y, iW, iH] do begin
              
              PaintingPixel(R1, G1, B1, cLeft, cTop);
              
              inc(cTop, mY);
              inc(cLeft, mX);
              
              if mY < 12 then inc(mY);
              
              if not mX = 0 then
                if mx < 0 then
                  inc(mX)
                else
                  dec(mX);
            end;
          end;
        end;
      end;
    end;
end;

procedure THITetris.RefreshBackGround;
var
  X, Y: word;
begin
  
  for X := 0 to FormW - 1 do
    for Y := 0 to FormH - 1 do
      TBig(FScr^)[X + Y * FormW] := RGB(GetRValue(FCanvaGamePanel[X, Y]),
        GetGValue(FCanvaGamePanel[X, Y]), GetBValue(FCanvaGamePanel[X, Y]));
end;

procedure THITetris.PaintClockFace(sNumber: string; cLeft, cHeight: Word;
  R, G, B: Byte; VisibleZero: Boolean);

  procedure PaintLineW(iLeft, iTop, iWidth: Word);
  var
    x: word;
  begin
    for x := iLeft to iLeft + iWidth do
      FCanvaGamePanel[x, iTop] := RGB(B, G, R);
  end;

  procedure PaintLineH(iLeft, iTop, iHeight: Word);
  var
    y: word;
  begin
    for y := iTop to iTop + iHeight do
      FCanvaGamePanel[iLeft, y] := RGB(B, G, R);
  end;

  procedure PaintUp(iLeft, iTop: Word);
  begin
    PaintLineW(iLeft + 1, iTop, 11);
    PaintLineW(iLeft + 2, iTop + 1, 9);
    PaintLineW(iLeft + 3, iTop + 2, 7);
  end;

  procedure PaintLeftUp(iLeft, iTop: Word);
  begin
    PaintLineH(iLeft, iTop + 1, 8);
    PaintLineH(iLeft + 1, iTop + 2, 7);
    PaintLineH(iLeft + 2, iTop + 3, 5);
  end;

  procedure PaintRightUp(iLeft, iTop: Word);
  begin
    PaintLineH(iLeft + 13, iTop + 1, 8);
    PaintLineH(iLeft + 12, iTop + 2, 7);
    PaintLineH(iLeft + 11, iTop + 3, 5);
  end;

  procedure PaintMid(iLeft, iTop: Word);
  begin
    PaintLineW(iLeft + 3, iTop + 9, 7);
    PaintLineW(iLeft + 2, iTop + 10, 9);
    PaintLineW(iLeft + 3, iTop + 11, 7);
  end;

  procedure PaintLeftDown(iLeft, iTop: Word);
  begin
    PaintLineH(iLeft, iTop + 11, 9);
    PaintLineH(iLeft + 1, iTop + 11, 8);
    PaintLineH(iLeft + 2, iTop + 12, 6);
  end;

  procedure PaintRightDown(iLeft, iTop: Word);
  begin
    PaintLineH(iLeft + 11, iTop + 12, 6);
    PaintLineH(iLeft + 12, iTop + 11, 8);
    PaintLineH(iLeft + 13, iTop + 11, 9);
  end;

  procedure PaintDown(iLeft, iTop: Word);
  begin
    PaintLineW(iLeft + 3, iTop + 19, 7);
    PaintLineW(iLeft + 2, iTop + 20, 9);
    PaintLineW(iLeft + 1, iTop + 21, 11);
  end;

var
  i, n: Byte;
  MyLeft: Word;
begin
  for i := 1 to 8 do begin
    
    MyLeft := cLeft + (i - 1) * 15;
    
    n := Str2Int(sNumber[i]);
    if (n = 0) and (VisibleZero = False) and (i <= 8 - Length(Int2Str(Str2Int(sNumber)))) then
    begin

    end else
    
      case n of
        0:
          begin
            PaintUp(MyLeft, cHeight);
            PaintLeftUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintLeftDown(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        1:
          begin
            PaintRightUp(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
          end;
        2:
          begin
            PaintUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintLeftDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        3:
          begin
            PaintUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        4:
          begin
            PaintLeftUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
          end;
        5:
          begin
            PaintUp(MyLeft, cHeight);
            PaintLeftUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        6:
          begin
            PaintUp(MyLeft, cHeight);
            PaintLeftUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintLeftDown(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        7:
          begin
            PaintUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
          end;
        8:
          begin
            PaintUp(MyLeft, cHeight);
            PaintLeftUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintLeftDown(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
        9:
          begin
            PaintUp(MyLeft, cHeight);
            PaintLeftUp(MyLeft, cHeight);
            PaintRightUp(MyLeft, cHeight);
            PaintMid(MyLeft, cHeight);
            PaintRightDown(MyLeft, cHeight);
            PaintDown(MyLeft, cHeight);
          end;
      end;
  end;
end;

procedure THITetris.PaintGrid;

  procedure NewGrid(cTop, cLeft, cHeight, cWidth, cMaxH, cMaxW: Word);
  var
    R, G, B: Byte;
    X, Y, I: word; 
    cMyLeft, cMyTop: Word;
  begin
    inc(cMaxW, -1); inc(cMaxH, -1);
    R := 138; G := R; B := R;
    
    for X := 0 to cMaxW do
    begin
        
      cMyLeft := cLeft + (PanelColorW + 1) * X;
        
      for I := cTop to cTop + cHeight - 1 do
        FCanvaGamePanel[cMyLeft, I] := RGB(B, G, R);
        
      if X = cMaxW then
          
        for I := cTop to cTop + cHeight - 1 do
          FCanvaGamePanel[cMyLeft + PanelColorW + 1, I] := RGB(B, G, R);
    end;
    
    for Y := 0 to cMaxH do
    begin
        
      cMyTop := cTop + (PanelColorH + 1) * Y;
        
      for I := cLeft to cLeft + cWidth - 1 do
        FCanvaGamePanel[I, cMyTop] := RGB(B, G, R);
      if Y = cMaxH then
          
        for I := cLeft to cLeft + cWidth - 1 do
          FCanvaGamePanel[I, cMyTop + (PanelColorH + 1)] := RGB(B, G, R);
    end;
  end;

var
  R, G, B: Byte; 
  cTop, cLeft, cTopV, cLeftV, cHeight, cWidth, cMaxH, cMaxW: Word;
begin
  
  cTop := FrameDepth - 1;
  cLeft := FrameDepth - 1;
  cHeight := (PanelColorH + 1) * PanelColorSummH;
  cWidth := (PanelColorW + 1) * PanelColorSummW;
  cMaxH := PanelColorSummH;
  cMaxW := PanelColorSummW;
  NewGrid(cTop, cLeft, cHeight, cWidth, cMaxH, cMaxW);
  
  cTop := FrameDepth + (PanelColorH + 1) * 9 - 1;
  cLeft := FrameDepth * 3 + (PanelColorW + 1) * (PanelColorSummW + 1) - 1;
  cHeight := (PanelColorH + 1) * 4;
  cWidth := (PanelColorW + 1) * 4;
  cMaxH := 4;
  cMaxW := 4;
  NewGrid(cTop, cLeft, cHeight, cWidth, cMaxH, cMaxW);
  
  cTop := FrameDepth + (PanelColorH + 1) * 15 - 1;
  cLeft := FrameDepth * 3 + (PanelColorW + 1) * (PanelColorSummW + 1) - 1;
  cHeight := (PanelColorH + 1) * 4;
  cWidth := (PanelColorW + 1) * 4;
  cMaxH := 4;
  cMaxW := 4;
  NewGrid(cTop, cLeft, cHeight, cWidth, cMaxH, cMaxW);
end;

procedure THITetris.FillBackGround;
var
  X, Y: word; 
  R, G, B: Byte; 
  PanelGameShiftX: word; 
begin
  
  R := 64; G := R; B := R;
  for Y := 0 to FrameDepth - 1 do
  begin
    for X := Y to FormW - Y - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  R := 64; G := R; B := R;
  for X := FormW - 1 downto FormW - FrameDepth do
  begin
    for Y := FormW - X to FormH - (FormW - X) - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  R := 64; G := R; B := R;
  for Y := FormH - 1 downto FormH - FrameDepth - 1 do
  begin
    for X := FormW - (FormH - Y) downto FormH - Y - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  R := 64; G := R; B := R;
  for X := 0 to FrameDepth - 1 do
  begin
    for Y := X to FormH - X - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  R := 64; G := R; B := R;
  PanelGameShiftX := PanelGameW + FrameDepth * 2;
  for X := PanelGameShiftX to PanelGameShiftX + FrameDepth - 1 do
  begin
    for Y := X - PanelGameShiftX to FormH - X + PanelGameShiftX - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  R := 64; G := R; B := R;
  PanelGameShiftX := PanelGameW + FrameDepth * 2;
  for X := PanelGameShiftX - 1 downto PanelGameShiftX - FrameDepth - 1 do
  begin
    for Y := PanelGameShiftX - X - 1 to FormH - (PanelGameShiftX - X) do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
    Inc(R, 3); Inc(G, 3); Inc(B, 3);
  end;
  
  PanelGameShiftX := PanelGameW + FrameDepth;
  for X := FrameDepth - 1 to PanelGameShiftX - 1 do
    for Y := FrameDepth - 1 to FormH - FrameDepth - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);
  PanelGameShiftX := PanelGameW + FrameDepth * 3;
  for X := PanelGameShiftX - 1 to FormW - FrameDepth - 1 do
    for Y := FrameDepth - 1 to FormH - FrameDepth - 1 do
      FCanvaGamePanel[X, Y] := RGB(B, G, R);

end;

procedure THITetris.SetLengthCanvaGamePanel;
begin
  SetLength(FCanvaGamePanel, FormW, FormH); 
  FillBackGround; 
  PaintGrid; 
end;

procedure THITetris.WriteText;
begin
  SetBkMode(FScrBitmap.Canvas.Handle, TRANSPARENT);
  FScrBitmap.Canvas.Font.FontHeight := 16;
  FScrBitmap.Canvas.TextOut(PanelGameW + FrameDepth * 2 + 36, 48, ':');
  SetBkMode(FScrBitmap.Canvas.Handle, OPAQUE);

  SetBkMode(FScrBitmap.Canvas.Handle, TRANSPARENT);
  FScrBitmap.Canvas.Font.FontHeight := 16;
  FScrBitmap.Canvas.TextOut(PanelGameW + FrameDepth * 2 + 36, 96, ':');
  SetBkMode(FScrBitmap.Canvas.Handle, OPAQUE);

  SetBkMode(FScrBitmap.Canvas.Handle, TRANSPARENT);
  FScrBitmap.Canvas.Font.FontHeight := 16;
  FScrBitmap.Canvas.TextOut(PanelGameW + FrameDepth * 2 + 36, 144, ':');
  SetBkMode(FScrBitmap.Canvas.Handle, OPAQUE);

  SetBkMode(FScrBitmap.Canvas.Handle, TRANSPARENT);
  FScrBitmap.Canvas.Font.FontHeight := 16;
  FScrBitmap.Canvas.TextOut(PanelGameW + FrameDepth * 2 + 36, 216, ':');
  SetBkMode(FScrBitmap.Canvas.Handle, OPAQUE);

  SetBkMode(FScrBitmap.Canvas.Handle, TRANSPARENT);
  FScrBitmap.Canvas.Font.FontHeight := 16;
  FScrBitmap.Canvas.TextOut(PanelGameW + FrameDepth * 2 + 36, 368, ':');
  SetBkMode(FScrBitmap.Canvas.Handle, OPAQUE);
end;

procedure THITetris.ReShowScore;
begin
  PaintClockFace('88888888', PanelGameW + FrameDepth * 2 + 36, 72, 108, 108, 108, True);
  PaintClockFace(FCurScore, PanelGameW + FrameDepth * 2 + 36, 72, 0, 0, 0, False);
  PaintClockFace('88888888', PanelGameW + FrameDepth * 2 + 36, 120, 108, 108, 108, True);
  PaintClockFace(FCurLevel, PanelGameW + FrameDepth * 2 + 36, 120, 0, 0, 0, False);
  PaintClockFace('88888888', PanelGameW + FrameDepth * 2 + 36, 168, 108, 108, 108, True);
  PaintClockFace(FCurLines, PanelGameW + FrameDepth * 2 + 36, 168, 0, 0, 0, False);
end;

procedure THITetris.TimerFPSTimer(Sender: PObj);

  procedure PainPanelInfo(txt: string);
  var
    x1, y1, x2, y2: Word;
  begin

    x1 := FrameDepth + PanelGameW div 32;
    x2 := FrameDepth + PanelGameW - PanelGameW div 32;
    y1 := FrameDepth + PanelGameH div 2 - 24;
    y2 := FrameDepth + PanelGameH div 2 + 240;

    FScrBitmap.Canvas.Pen.Color := RGB(0, 0, 0);
    FScrBitmap.Canvas.Brush.Color := RGB(120, 120, 120);
    FScrBitmap.Canvas.Pen.PenWidth := 1;
    FScrBitmap.Canvas.Rectangle(x1 + 1, y1 + 1, x2 - 1, y2 - 1);

    FScrBitmap.Canvas.Font.FontHeight := 16;
    FScrBitmap.Canvas.TextOut(x1 + 16, y1 + 8, txt);

    FScrBitmap.Canvas.Font.FontHeight := 12;
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 60, ' - / ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 80, 'Z -   -90 ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 100, 'X -  ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 120, 'C -   90 ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 140, 'P - /');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 160, ' -  ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 180, ' -   ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 200, ' -  ');
    FScrBitmap.Canvas.TextOut(x1 + 8, y1 + 220, ' -  ');
  end;

begin
  RefreshBackGround; 
  PaintingAllPanels; 
  ReShowScore; 
  WriteText; 
  
  if FGameOver then
  begin
    
    PainPanelInfo(' ...');
  end else if FCurrGamePause then
  begin
    
    PainPanelInfo('');
  end;
  FormPaint(Control, Control.Canvas.Handle); 
end;

procedure THITetris.TimerGravityTimer(Sender: PObj);
begin
  if not FGameOver and not FCurrGamePause then
    MoweObjectColor(0);
end;

procedure THITetris.CreateScrBitMap(BitMapH, BitMapW: Integer);
var
  BInfo: tagBITMAPINFO;
begin
  BInfo.bmiHeader.biSize := sizeof(tagBITMAPINFOHEADER);
  BInfo.bmiHeader.biWidth := BitMapH;
  BInfo.bmiHeader.biHeight := -BitMapW;
  BInfo.bmiHeader.biPlanes := 1;
  BInfo.bmiHeader.biBitCount := 32;
  BInfo.bmiHeader.biCompression := BI_RGB;
  FScrBitmap := NewBitmap(1, 1);
  FScrBitmap.Handle := CreateDIBSection(Control.Canvas.Handle, BInfo, DIB_RGB_COLORS,
    FScr, 0, 0);
  ZeroMemory(FScr, BitMapH * BitMapW * 4);
end;

procedure THITetris.SetLengthFormPanelColor;
var
  x, y: Byte;
begin
  if not IsInit then
  begin
    
    SetLength(FormPanelColor, PanelColorSummW, PanelColorSummH);
    IsInit := True;
  end;
  
  for x := 0 to PanelColorSummW - 1 do
    for y := 0 to PanelColorSummH - 1 do
      with FormPanelColor[x, y] do begin
        Visible := False;
        with PanelColor do begin
          R1 := 0; R2 := 0;
          G1 := 0; G2 := 0;
          B1 := 0; B2 := 0;
        end;
      end;
end;

procedure THITetris.Init;
begin
  Control := _NewControl(FParent, 'THITetris', WS_VISIBLE or WS_CHILD or WS_TABSTOP, False, nil);
  inherited;
  IsInit := False;
  FGameOver := True;
  FCurScore := '00000000';
  FCurLevel := '00000001';
  FCurLines := '00000000';
  yNumLines := 0;
  FCurrGamePause := False;
  
  Control.ClientHeight := FormH;
  Control.ClientWidth := FormW;
  
  CreateScrBitmap(Control.ClientWidth, Control.ClientHeight);
  
  SetLengthCanvaGamePanel;
  
  RefreshBackGround;
  
  SetLengthFormPanelColor;
  
  SetLength(PixelToClear, PanelColorSummW, 4, PanelColorW, PanelColorH);

  Control.OnPaint := FormPaint;

  TimerFPS := NewTimer(40); 
  TimerFPS.Enabled := True;
  TimerFPS.OnTimer := TimerFPSTimer;

  TimerGravity := NewTimer(800); 
  TimerGravity.Enabled := True;
  TimerGravity.OnTimer := TimerGravityTimer;

end;

procedure THITetris.ReStartGame;
begin
  FCurScore := '00000000';
  FCurLevel := '00000001';
  FCurLines := '00000000';
  FCurrGamePause := False;
  
  ObjectColorCurr.Visible := False; 
  ObjectColorBuff.Visible := False; 
  ObjectColorCapt.Visible := False; 
  ObjectColorShad.Visible := False; 
  ObjectColorNext.Visible := False; 
  
  SetLengthFormPanelColor;
  
  CreateNextObjectColor(random(7));
  
  ObjectColorCurr := ObjectColorNext;
  
  CreateNextObjectColor(random(7));
  
  FGameOver := False;
end;

procedure THITetris.ChangeObjectColor;
begin
  
  ObjectColorBuff := ObjectColorCurr;
  
  if ObjectColorCapt.Visible then
  begin
    
    ObjectColorCurr := ObjectColorCapt;
    
    ObjectColorCapt := ObjectColorBuff;
  end else
  begin
    
    ObjectColorCurr := ObjectColorNext;
    
    CreateNextObjectColor(random(7));
  end;
  ObjectColorCapt := ObjectColorBuff;
  with ObjectColorCapt do begin
    Top := -Height;
    Left := (PanelColorSummW - Width) div 2;
  end;
end;

procedure THITetris._onKeyDown(Sender: PControl; var Key: Longint; Shift: DWORD);
begin
  case key of
    32: ReStartGame; 
    37: if not FGameOver and not FCurrGamePause then MoweObjectColor(1); 
    39: if not FGameOver and not FCurrGamePause then MoweObjectColor(2); 
    40: if not FGameOver and not FCurrGamePause then
      begin
        MoweObjectColor(3); 
        FCurScore := IncScore(1, FCurScore); 
      end;
    38: if not FGameOver and not FCurrGamePause then MoweObjectColor(4); 
    88: if not FGameOver and not FCurrGamePause then
      begin
        ChangeObjectColor; 
        MoweObjectColor(5);
      end;
    67: if not FGameOver and not FCurrGamePause then TurnObjectColor(90);
    90: if not FGameOver and not FCurrGamePause then TurnObjectColor(-90);
    80: if not FGameOver then FCurrGamePause := not FCurrGamePause;
  end;
end;

procedure THITetris.FormPaint(Sender: PControl; DC: HDC);
begin
  FScrBitmap.Draw(Control.Canvas.Handle, 0, 0);
end;

end.


